print("import: 'Bio'")
import Bio

print("import: 'Bio.Affy'")
import Bio.Affy

print("import: 'Bio.Align'")
import Bio.Align

print("import: 'Bio.Align.Applications'")
import Bio.Align.Applications

print("import: 'Bio.AlignIO'")
import Bio.AlignIO

print("import: 'Bio.Alphabet'")
import Bio.Alphabet

print("import: 'Bio.Application'")
import Bio.Application

print("import: 'Bio.Blast'")
import Bio.Blast

print("import: 'Bio.CAPS'")
import Bio.CAPS

print("import: 'Bio.Cluster'")
import Bio.Cluster

print("import: 'Bio.Compass'")
import Bio.Compass

print("import: 'Bio.Crystal'")
import Bio.Crystal

print("import: 'Bio.Data'")
import Bio.Data

print("import: 'Bio.Emboss'")
import Bio.Emboss

print("import: 'Bio.Entrez'")
import Bio.Entrez

print("import: 'Bio.ExPASy'")
import Bio.ExPASy

print("import: 'Bio.FSSP'")
import Bio.FSSP

print("import: 'Bio.GenBank'")
import Bio.GenBank

print("import: 'Bio.Geo'")
import Bio.Geo

print("import: 'Bio.HMM'")
import Bio.HMM

print("import: 'Bio.KDTree'")
import Bio.KDTree

print("import: 'Bio.KEGG'")
import Bio.KEGG

print("import: 'Bio.KEGG.Compound'")
import Bio.KEGG.Compound

print("import: 'Bio.KEGG.Enzyme'")
import Bio.KEGG.Enzyme

print("import: 'Bio.KEGG.KGML'")
import Bio.KEGG.KGML

print("import: 'Bio.KEGG.Map'")
import Bio.KEGG.Map

print("import: 'Bio.Medline'")
import Bio.Medline

print("import: 'Bio.NMR'")
import Bio.NMR

print("import: 'Bio.Nexus'")
import Bio.Nexus

print("import: 'Bio.PDB'")
import Bio.PDB

print("import: 'Bio.PDB.QCPSuperimposer'")
import Bio.PDB.QCPSuperimposer

print("import: 'Bio.Pathway'")
import Bio.Pathway

print("import: 'Bio.Pathway.Rep'")
import Bio.Pathway.Rep

print("import: 'Bio.Phylo'")
import Bio.Phylo

print("import: 'Bio.Phylo.Applications'")
import Bio.Phylo.Applications

print("import: 'Bio.Phylo.PAML'")
import Bio.Phylo.PAML

print("import: 'Bio.PopGen'")
import Bio.PopGen

print("import: 'Bio.PopGen.GenePop'")
import Bio.PopGen.GenePop

print("import: 'Bio.Restriction'")
import Bio.Restriction

print("import: 'Bio.SCOP'")
import Bio.SCOP

print("import: 'Bio.SVDSuperimposer'")
import Bio.SVDSuperimposer

print("import: 'Bio.SearchIO'")
import Bio.SearchIO

print("import: 'Bio.SearchIO.BlastIO'")
import Bio.SearchIO.BlastIO

print("import: 'Bio.SearchIO.ExonerateIO'")
import Bio.SearchIO.ExonerateIO

print("import: 'Bio.SearchIO.HmmerIO'")
import Bio.SearchIO.HmmerIO

print("import: 'Bio.SearchIO._model'")
import Bio.SearchIO._model

print("import: 'Bio.SeqIO'")
import Bio.SeqIO

print("import: 'Bio.SeqUtils'")
import Bio.SeqUtils

print("import: 'Bio.Sequencing'")
import Bio.Sequencing

print("import: 'Bio.Sequencing.Applications'")
import Bio.Sequencing.Applications

print("import: 'Bio.Statistics'")
import Bio.Statistics

print("import: 'Bio.SubsMat'")
import Bio.SubsMat

print("import: 'Bio.SwissProt'")
import Bio.SwissProt

print("import: 'Bio.TogoWS'")
import Bio.TogoWS

print("import: 'Bio.UniGene'")
import Bio.UniGene

print("import: 'Bio.UniProt'")
import Bio.UniProt

print("import: 'Bio.Wise'")
import Bio.Wise

print("import: 'Bio._py3k'")
import Bio._py3k

print("import: 'Bio.codonalign'")
import Bio.codonalign

print("import: 'Bio.cpairwise2'")
import Bio.cpairwise2

print("import: 'Bio.motifs'")
import Bio.motifs

print("import: 'Bio.motifs.applications'")
import Bio.motifs.applications

print("import: 'Bio.motifs.jaspar'")
import Bio.motifs.jaspar

print("import: 'Bio.phenotype'")
import Bio.phenotype

print("import: 'Bio.trie'")
import Bio.trie

print("import: 'BioSQL'")
import BioSQL

